# Code borrowed from https://github.com/danieldeutsch/qaeval/tree/master
import collections
import edlib
from typing import Tuple

from typing import List, Dict, Set, Tuple
from transformers import AutoModelForSequenceClassification, AutoTokenizer
from transformers.data.metrics.squad_metrics import (
    get_final_text,
    _get_best_indexes,
    _compute_softmax,
    compute_exact,
    compute_f1
)

def _is_whitespace(c):
    if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
        return True
    return False

def _get_char_offsets(example, pred_start, pred_end):
    # The returned end index will be exclusive
    if pred_start is None or pred_end is None:
        # This could happen if there's an edge case with no valid predictions. See where the prediction is "empty"
        return None, None

    token_to_char_start = {}
    token_to_char_end = {}
    for char_index, token_index in enumerate(example.char_to_word_offset):
        if token_index not in token_to_char_start:
            token_to_char_start[token_index] = char_index
        token_to_char_end[token_index] = char_index

    # Any whitespace after the token is included in that token. Find the last non-whitespace character
    for token_index, end in token_to_char_end.items():
        if token_index == -1:
            # Whitespace at the beginning is mapped to token -1. We don't care about it
            continue
        while _is_whitespace(example.context_text[end]):
            end -= 1
            if end < 0:
                break
        if end < 0:
            raise Exception(f'Token end is less than 0.')
        token_to_char_end[token_index] = end + 1  # exclusive
    return token_to_char_start[pred_start], token_to_char_end[pred_end]


def compute_predictions_logits_with_null(
        tokenizer,
        all_examples,
        all_features,
        all_results,
        n_best_size,
        max_answer_length,
        do_lower_case,
        verbose_logging,
        version_2_with_negative,
        return_offsets = False
):
    example_index_to_features = collections.defaultdict(list)
    for feature in all_features:
        example_index_to_features[feature.example_index].append(feature)

    unique_id_to_result = {}
    for result in all_results:
        unique_id_to_result[result.unique_id] = result

    _PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
        "PrelimPrediction", ["feature_index", "start_index", "end_index", "start_logit", "end_logit"]
    )

    all_predictions = collections.OrderedDict()
    all_nbest_json = collections.OrderedDict()
    scores_diff_json = collections.OrderedDict()
    all_probs = collections.OrderedDict()
    null_scores = collections.OrderedDict()
    offsets = collections.OrderedDict()

    for (example_index, example) in enumerate(all_examples):
        features = example_index_to_features[example_index]

        prelim_predictions = []
        # keep track of the minimum score of null start+end of position 0
        score_null = 1000000  # large and positive
        min_null_feature_index = 0  # the paragraph slice with min null score
        null_start_logit = 0  # the start logit at the slice with min null score
        null_end_logit = 0  # the end logit at the slice with min null score
        for (feature_index, feature) in enumerate(features):
            result = unique_id_to_result[feature.unique_id]
            start_indexes = _get_best_indexes(result.start_logits[0], n_best_size)
            end_indexes = _get_best_indexes(result.end_logits[0], n_best_size)
            # if we could have irrelevant answers, get the min score of irrelevant
            if version_2_with_negative:
                # TODO 
                feature_null_score = result.start_logits[0][0] + result.end_logits[0][0]
                if feature_null_score < score_null:
                    score_null = feature_null_score
                    min_null_feature_index = feature_index
                    null_start_logit = result.start_logits[0][0]
                    null_end_logit = result.end_logits[0][0]

            for start_index in start_indexes:
                for end_index in end_indexes:
                    # We could hypothetically create invalid predictions, e.g., predict
                    # that the start of the span is in the question. We throw out all
                    # invalid predictions.
                    if start_index >= len(feature.tokens):
                        continue
                    if end_index >= len(feature.tokens):
                        continue
                    if start_index not in feature.token_to_orig_map:
                        continue
                    if end_index not in feature.token_to_orig_map:
                        continue
                    if not feature.token_is_max_context.get(start_index, False):
                        continue
                    if end_index < start_index:
                        continue
                    length = end_index - start_index + 1
                    if length > max_answer_length:
                        continue
                    prelim_predictions.append(
                        _PrelimPrediction(
                            feature_index=feature_index,
                            start_index=start_index,
                            end_index=end_index,
                            start_logit=result.start_logits[0][start_index],
                            end_logit=result.end_logits[0][end_index],
                        )
                    )
        if version_2_with_negative:
            prelim_predictions.append(
                _PrelimPrediction(
                    feature_index=min_null_feature_index,
                    start_index=0,
                    end_index=0,
                    start_logit=null_start_logit,
                    end_logit=null_end_logit,
                )
            )

        prelim_predictions = sorted(prelim_predictions, key=lambda x: (x.start_logit + x.end_logit), reverse=True)

        _NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "NbestPrediction", ["text", "start_logit", "end_logit", "doc_start", "doc_end"]
        )

        seen_predictions = {}
        nbest = []
        for pred in prelim_predictions:
            if len(nbest) >= n_best_size:
                break
            feature = features[pred.feature_index]
            if pred.start_index > 0:  # this is a non-null prediction
                tok_tokens = feature.tokens[pred.start_index: (pred.end_index + 1)]
                orig_doc_start = feature.token_to_orig_map[pred.start_index]
                orig_doc_end = feature.token_to_orig_map[pred.end_index]
                orig_tokens = example.doc_tokens[orig_doc_start: (orig_doc_end + 1)]

                tok_text = tokenizer.convert_tokens_to_string(tok_tokens)

                # tok_text = " ".join(tok_tokens)
                #
                # # De-tokenize WordPieces that have been split off.
                # tok_text = tok_text.replace(" ##", "")
                # tok_text = tok_text.replace("##", "")

                # Clean whitespace
                tok_text = tok_text.strip()
                tok_text = " ".join(tok_text.split())
                orig_text = " ".join(orig_tokens)

                final_text = get_final_text(tok_text, orig_text, do_lower_case, verbose_logging)
                if final_text in seen_predictions:
                    continue

                seen_predictions[final_text] = True
            else:
                final_text = ""
                orig_doc_start = None
                orig_doc_end = None
                seen_predictions[final_text] = True

            nbest.append(_NbestPrediction(text=final_text, start_logit=pred.start_logit, end_logit=pred.end_logit,
                                          doc_start=orig_doc_start, doc_end=orig_doc_end))
        # if we didn't include the empty option in the n-best, include it
        if version_2_with_negative:
            if "" not in seen_predictions:
                nbest.append(_NbestPrediction(text="", start_logit=null_start_logit, end_logit=null_end_logit,
                                              doc_start=None, doc_end=None))

            # In very rare edge cases we could only have single null prediction.
            # So we just create a nonce prediction in this case to avoid failure.
            if len(nbest) == 1:
                nbest.insert(0, _NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0, doc_start=None,
                                                 doc_end=None))

        # In very rare edge cases we could have no valid predictions. So we
        # just create a nonce prediction in this case to avoid failure.
        if not nbest:
            nbest.append(_NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0, doc_start=None,
                                          doc_end=None))

        assert len(nbest) >= 1

        total_scores = []
        best_non_null_entry = None
        best_non_null_entry_index = None
        for i, entry in enumerate(nbest):
            total_scores.append(entry.start_logit + entry.end_logit)
            if not best_non_null_entry:
                if entry.text:
                    best_non_null_entry = entry
                    best_non_null_entry_index = i

        probs = _compute_softmax(total_scores)

        nbest_json = []
        null_prob = None
        best_prob = None
        for (i, entry) in enumerate(nbest):
            output = collections.OrderedDict()
            output["text"] = entry.text
            output["probability"] = probs[i]
            output["start_logit"] = entry.start_logit
            output["end_logit"] = entry.end_logit
            if entry.text == '':
                null_prob = probs[i]
            if i == best_non_null_entry_index:
                best_prob = probs[i]
            nbest_json.append(output)

        assert len(nbest_json) >= 1

        if not version_2_with_negative:
            all_predictions[example.qas_id] = nbest_json[0]["text"]
        else:
            # Always predict the best non-null text
            all_predictions[example.qas_id] = best_non_null_entry.text
            all_probs[example.qas_id] = best_prob
            null_scores[example.qas_id] = null_prob
            offsets[example.qas_id] = _get_char_offsets(example, best_non_null_entry.doc_start,
                                                        best_non_null_entry.doc_end)

            # # predict "" iff the null score - the score of best non-null > threshold
            # score_diff = score_null - best_non_null_entry.start_logit - (best_non_null_entry.end_logit)
            # scores_diff_json[example.qas_id] = score_diff
            # if score_diff > null_score_diff_threshold:
            #     all_predictions[example.qas_id] = ""
            # else:
            #     all_predictions[example.qas_id] = best_non_null_entry.text
        all_nbest_json[example.qas_id] = nbest_json

    output = (all_predictions, all_probs, null_scores)
    if return_offsets:
        output = output + (offsets,)
    return output


class SpanFixError(Exception):
    pass


def fix_answer_span(prediction: str, document_span: str, start: int, end: int) -> Tuple[int, int]:
    """
    Tries to fix the answer span of the prediction, which may include some extra whitespace or special characters.

    # Parameters
    - `prediction`: the string output by the QA model
    - `document_span`: the span in the text given by the maybe noisy offsets. See `QuestionAnsweringModel.answer()`
    documentation for more information
    - `start`: the start character offset of `document_span` in the original text
    - `end`: the *exclusive* end character offset of the `document_span` in the original text

    # Returns
    The `start` and *exclusive* `end` character offsets of fixed character offsets of `prediction` in the
    original text.
    """

    if len(prediction) > len(document_span):
        raise SpanFixError(f'Unexpected lengths: "{prediction}", "{document_span}"')

    alignment = edlib.align(prediction, document_span, mode='HW', task='path')
    locations = alignment['locations']
    if len(locations) != 1:
        raise SpanFixError(f'Unable to compute span: "{prediction}", "{document_span}"')
    align_start, align_end = locations[0]

    start += align_start
    end -= len(document_span) - align_end
    end += 1
    return start, end



class Scorer(object):
    def keys(self) -> Set[str]:
        raise NotImplementedError

    def default_scores(self) -> Dict[str, float]:
        return {key: 0.0 for key in self.keys()}

    def score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> Tuple[Dict[str, float], List[Dict[str, float]]]:
        scores_dicts = self._score_single_ref(
            context,
            questions,
            answers,
            predictions,
            probabilities,
            null_probabilities
        )
        aggregated_scores = self.aggregate_scores(scores_dicts)
        return aggregated_scores, scores_dicts

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        raise NotImplementedError

    def score_multi_ref(
        self,
        context: str,
        questions_list: List[List[str]],
        answers_list: List[List[str]],
        predictions_list: List[List[str]],
        probabilities_list: List[List[float]],
        null_probabilities_list: List[List[float]]
    ) -> Tuple[Dict[str, float], List[List[Dict[str, float]]]]:
        # The aggregated per-reference scores
        reference_scores_list = []
        # The scores for each individual question. [i][j] will be the scores from
        # reference i and question j
        question_scores_list = []

        for i in range(len(questions_list)):
            reference_scores, question_scores = self.score_single_ref(
                context,
                questions_list[i],
                answers_list[i],
                predictions_list[i],
                probabilities_list[i],
                null_probabilities_list[i]
            )
            reference_scores_list.append(reference_scores)
            question_scores_list.append(question_scores)

        instance_scores = self.aggregate_scores(reference_scores_list)
        return instance_scores, question_scores_list

    def _ensure_expected_keys(self, expected_keys: Set[str], scores_dicts: List[Dict[str, float]]) -> None:
        for scores in scores_dicts:
            if expected_keys != scores.keys():
                raise Exception(f'Unequal keys: {expected_keys}; {scores.keys()}')

    def aggregate_scores(self, scores_dicts: List[Dict[str, float]]) -> Dict[str, float]:
        if len(scores_dicts) == 0:
            return self.default_scores()

        expected_keys = self.keys()
        self._ensure_expected_keys(expected_keys, scores_dicts)
        sums = {key: 0.0 for key in expected_keys}
        for scores in scores_dicts:
            for key in expected_keys:
                sums[key] += scores[key]

        averages = {key: sums[key] / len(scores_dicts) for key in expected_keys}
        return averages


def predict(our_model, our_tokenizer, batch_sentences, device):
    inputs = our_tokenizer(batch_sentences, max_length=512,  truncation=True, \
        padding="max_length", return_tensors="pt")
    outputs = our_model(input_ids=inputs["input_ids"].to(device), \
        attention_mask=inputs["attention_mask"].to(device))
    outputs = [x[0] for x in outputs[0].cpu().tolist()]
    outputs = [{"pred_score": x} for x in outputs]

    return outputs


class LERCQuipScorer(Scorer):
    def __init__(self, lerc_quip_path: str, cuda_device: int, batch_size: int = 8) -> None:
        self.device = cuda_device

        self.predictor = AutoModelForSequenceClassification.from_pretrained(lerc_quip_path)
        if self.device >= 0:
            self.predictor.to(self.device)
        self.predictor.eval()

        self.tokenizer = AutoTokenizer.from_pretrained(lerc_quip_path)
        self.batch_size = batch_size

    def keys(self) -> Set[str]:
        return {'lerc_quip'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        input_dicts = []
        indices = []
        for i, (answer, question, prediction, probability, null_probability) in \
                enumerate(zip(answers, questions, predictions,
                    probabilities, null_probabilities)):
            if probability > null_probability:
                sentence1 = f"{question} <q> {answer} <r> {prediction} <c> {context}"
                input_dicts.append(sentence1)
                indices.append(i)

        output_dicts = []
        for i in range(0, len(input_dicts), self.batch_size):
            batch = input_dicts[i:i + self.batch_size]
            output_dicts.extend(predict(self.predictor, self.tokenizer, batch, self.device))
        assert len(output_dicts) == len(input_dicts)

        scores = [0.0] * len(questions)
        for i, output_dict in zip(indices, output_dicts):
            scores[i] = output_dict['pred_score']
        scores = [{'lerc_quip': s} for s in scores]
        return scores


class MetaScorer(Scorer):
    def __init__(self, scorers: List['Scorer']) -> None:
        self.scorers = scorers

    def _merge_dicts(self, dicts: List[Dict[str, float]]) -> Dict[str, float]:
        merged = {}
        for other in dicts:
            merged.update(other)
        return merged

    def keys(self) -> Set[str]:
        keys = set()
        for scorer in self.scorers:
            keys |= scorer.keys()
        return keys

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores_list = []
        for scorer in self.scorers:
            _, scores = scorer.score_single_ref(
                context,
                questions,
                answers,
                predictions,
                probabilities,
                null_probabilities
            )
            scores_list.append(scores)

        combined_scores = []
        for i in range(len(questions)):
            combined_scores.append(self._merge_dicts([scores[i] for scores in scores_list]))
        return combined_scores
    


class F1Scorer(Scorer):
    def keys(self) -> Set[str]:
        return {'f1'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores = []
        for prediction, answer, prob, null_prob in zip(predictions, answers, probabilities, null_probabilities):
            if prediction is None or null_prob >= prob:
                scores.append({'f1': 0.0})
            else:
                scores.append({'f1': compute_f1(answer, prediction)})
        return scores

class IsAnsweredScorer(Scorer):
    def keys(self) -> Set[str]:
        return {'is_answered'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores = []
        for prob, null_prob in zip(probabilities, null_probabilities):
            if prob > null_prob:
                scores.append({'is_answered': 1.0})
            else:
                scores.append({'is_answered': 0.0})
        return scores


class ExactMatchScorer(Scorer):
    def keys(self) -> Set[str]:
        return {'em'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores = []
        for prediction, answer, prob, null_prob in zip(predictions, answers, probabilities, null_probabilities):
            if prediction is None or null_prob >= prob:
                scores.append({'em': 0.0})
            else:
                scores.append({'em': compute_exact(answer, prediction)})
        return scores


'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

import nltk
from rank_bm25 import BM25Okapi
from nltk.corpus import stopwords
import numpy as np


def get_instruction(sample):
    if sample["prompt_all"] == "":
        return ' '.join([sample["prompt_task"], sample["prompt_context"]])
    return sample["prompt_all"]


def preprocess_knowledge(text, english_stopwords):
    tokens = nltk.word_tokenize(text.lower())
    return [token for token in tokens if token not in english_stopwords]


def retrieve_global_knowledge(output, global_knowledge, top_n=5):
    if len(global_knowledge) < top_n:
        return '\n'.join(global_knowledge)
    english_stopwords = set(stopwords.words('english'))
    knowledge = [preprocess_knowledge(paragraph, english_stopwords) for paragraph in global_knowledge]
    bm25 = BM25Okapi(knowledge)
    query = preprocess_knowledge(output, english_stopwords)
    retrieved_knowledge = bm25.get_top_n(query, global_knowledge, n=top_n)  # For now, retrieving top 5
    return '\n'.join(retrieved_knowledge)



'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

from langchain import PromptTemplate

template = """Conversation history:

{prompt}
Response: {response}
Scores:
{scores}

Use the feedback scores to improve the response.

Conversation history:

{prompt}
Response:"""

prompt_template = PromptTemplate(input_variables=["prompt", "response", "scores"], template=template)



'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

from .constants import prompt_template

def refine(prompt, response, feedback_scores, model):
    scores = [f"{key}: {score}" for key, score in feedback_scores.items() if key.endswith("_score")]
    input_txt = prompt_template.format(prompt=prompt, response=response, scores=scores)
    gen_param = {
        "max_tokens":128
    }
    return model.generate(input_txt, **gen_param)



'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

from langchain import PromptTemplate

request_factuality_guidelines_template = """Instruction: {model_input}

Output: {model_output}

The model-generated output may contain facts that are inconsistent with the facts in the instruction. To ensure the model's output is accurate and does not contain any false information, create guidelines that require the model to reproduce all facts from the instruction exactly as they are stated.

Guidelines:
"""

request_factuality_guidelines_prompt_template = PromptTemplate(input_variables=["model_input", "model_output"], template=request_factuality_guidelines_template)

revised_factuality_template = """Instruction: {model_input}

Follow the guidelines below:
{guideline}

"""

revised_factuality_prompt_template = PromptTemplate(input_variables=["model_input", "guideline"], template=revised_factuality_template)



# This code does not work as is, but was kept for reference, if others are interested in re-running baselines.
from utils_summac_benchmark import load_factcc, load_polytope, load_cogensumm, load_frank, load_summeval, load_xsumfaith
import sklearn, numpy as np, os, pandas as pd, sys, argparse
from model_baseline import BaselineScorer
from utils_scoring import ScorerWrapper
import utils_misc, seaborn as sns
from collections import Counter

sys.path.insert(0, "/home/phillab/feqa/")
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

# For now, can't use argparse because it is hard-coded in DAE... very shitty
# parser = argparse.ArgumentParser()
# parser.add_argument("--model", type=str, choices=["questeval", "feqa", "dae"], default="questeval")
# parser.add_argument("--questeval_weighter", action="store_true")
# parser.add_argument("--cut", type=str, choices=["val", "test"], default="val")
# args = parser.parse_args()

model = "dae"
cut = "test"

utils_misc.DoublePrint("%s_%s.log" % (model, cut))

def choose_best_threshold(labels, scores):
    best_bacc = 0.0
    best_thresh = 0.0
    thresholds = [np.percentile(scores, p) for p in np.arange(0, 100, 1)]
    for thresh in thresholds:
        preds = [1 if score > thresh else 0 for score in scores]
        bacc_score = sklearn.metrics.balanced_accuracy_score(labels, preds)
        if bacc_score >= best_bacc:
            best_bacc = bacc_score
            best_thresh = thresh
    return best_thresh

def from_score_to_pred(dataset, score_key):
    scores = [d[score_key] for d in dataset]
    labels = [d["label"] for d in dataset]
    thresh = choose_best_threshold(labels, scores)

    pred_key = "pred_%s" % (score_key)
    for d in dataset:
        d[pred_key] = 1 if d[score_key]>thresh else 0


datasets = [{"name": "factcc", "dataset": load_factcc(cut=cut)},
            {"name": "frank", "dataset": load_frank(cut=cut)},
            {"name": "pt_any", "dataset": load_polytope(which_label="overall", cut=cut)},
            {"name": "summ_corr", "dataset": load_cogensumm(cut=cut)},
            {"name": "summeval", "dataset": load_summeval(cut=cut)},
            {"name": "xsumfaith", "dataset": load_xsumfaith(cut=cut)}
            ]

dataset_stats = []
for dataset in datasets:
    N_pos, N_neg = len([d for d in dataset["dataset"] if d["label"]==1]), len([d for d in dataset["dataset"] if d["label"]==0])
    dataset_stats.append({"name": dataset["name"], "N": len(dataset["dataset"]), "N_pos": N_pos, "N_neg": N_neg, "frac_pos": N_pos/(N_pos+N_neg)})

print(pd.DataFrame(dataset_stats))

scorers = []
if model == "questeval":
    scorers.append({"name": "QuestEval", "model": BaselineScorer(model="questeval", do_weighter=args.questeval_weighter), "sign": 1})
elif model == "feqa":
    scorers.append({"name": "FEQA", "model": BaselineScorer(model="feqa"), "sign": 1})
elif model == "dae":
    scorers.append({"name": "DAE", "model": BaselineScorer(model="dae"), "sign": 1})


for scorer in scorers:
    scorer["model"].load_cache()

batch_size = 100
scorer_doc = ScorerWrapper(scorers, scoring_method="sum", max_batch_size=batch_size, use_caching=True)

def compute_doc_level(dataset):
    documents = [d["document"] for d in dataset]
    summaries = [d["claim"] for d in dataset]
    doc_scores = scorer_doc(documents, summaries, progress=True)
    label_keys = [k for k in doc_scores.keys() if "_scores" in k]

    for label_key in label_keys:
        score_key = ("%s|doc" % (label_key)).replace("_scores", "")
        for d, score in zip(dataset, doc_scores[label_key]):
            d[score_key] = score
        from_score_to_pred(dataset, score_key)


results = []
for dataset in datasets:
    print("======= %s ========" % (dataset["name"]))
    datas = dataset["dataset"]
    compute_doc_level(datas)
    for scorer in scorers:
        scorer["model"].save_cache()

    pred_labels = [k for k in datas[0].keys() if "pred_" in k]
    for pred_label in pred_labels:
        preds = [d[pred_label] for d in datas]
        labels = [d["label"] for d in datas]
        model_name, input_type = pred_label.replace("pred_", "").split("|")

        label_counts = Counter([d["label"] for d in dataset["dataset"]])
        pos_label = 0 if label_counts[0] < label_counts[1] else 1

        f1 = sklearn.metrics.f1_score(labels, preds, pos_label=pos_label)
        balanced_acc = sklearn.metrics.balanced_accuracy_score(labels, preds)

        results.append({"model_name": model_name, "dataset_name": dataset["name"], "input": input_type, "%s_f1" % (dataset["name"]): f1, "%s_bacc" % (dataset["name"]): balanced_acc})

cm = sns.light_palette("green", as_cmap=True)

def highlight_max(data):
    is_max = data == data.max()
    return ['font-weight: bold' if v else '' for v in is_max]


df = pd.DataFrame(results)
df = df.groupby(["model_name", "input"]).agg({"%s_bacc" % (d["name"]): "mean" for d in datasets})

df = df.rename(columns={k: k.replace("_bacc", "") for k in df.keys()})
df.drop("total", inplace=True)

print(df)
df.to_csv("/home/phillab/%s_results.csv" % (model))
# print(df.style.apply(highlight_max).background_gradient(cmap=cm, high=1.0, low=0.0).set_precision(3).set_caption("Weighed Accuracy"))



# borrowed from https://github.com/maszhongming/UniEval/blob/main/metric/scorer.py

import torch
import torch.nn as nn
from transformers import AutoConfig, AutoTokenizer, AutoModelForSeq2SeqLM
from tqdm import tqdm

class UniEvaluator:
    def __init__(self, model_name_or_path, max_length=1024, device='cuda:0', cache_dir=None):
        """ Set up model """
        self.device = device
        self.max_length = max_length

        self.config = AutoConfig.from_pretrained(model_name_or_path, cache_dir=cache_dir)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, cache_dir=cache_dir)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name_or_path, config=self.config,
                                                           cache_dir=cache_dir)

        self.model.eval()
        self.model.to(device)

        self.softmax = nn.Softmax(dim=1)

        self.pos_id = self.tokenizer("Yes")["input_ids"][0]
        self.neg_id = self.tokenizer("No")["input_ids"][0]

    def score(self, inputs, batch_size=8):
        """
            Get scores for the given samples.
            final_score = postive_score / (postive_score + negative_score)
        """

        # The implementation of "forward" in T5 still requires decoder_input_ids.
        # Therefore, we construct a random one-word target sequence.
        # The content of the target has no effect on the final scores.
        tgts = ["No" for _ in range(len(inputs))]

        pos_score_list, neg_score_list = [], []
        for i in tqdm(range(0, len(inputs), batch_size)):
            src_list = inputs[i: i + batch_size]
            tgt_list = tgts[i: i + batch_size]
            try:
                with torch.no_grad():
                    encoded_src = self.tokenizer(
                        src_list,
                        max_length=self.max_length,
                        truncation=True,
                        padding=True,
                        return_tensors='pt'
                    )
                    encoded_tgt = self.tokenizer(
                        tgt_list,
                        max_length=self.max_length,
                        truncation=True,
                        padding=True,
                        return_tensors='pt'
                    )

                    src_tokens = encoded_src['input_ids'].to(self.device)
                    src_mask = encoded_src['attention_mask'].to(self.device)

                    tgt_tokens = encoded_tgt['input_ids'].to(self.device)[:, 0].unsqueeze(-1)

                    output = self.model(
                        input_ids=src_tokens,
                        attention_mask=src_mask,
                        labels = tgt_tokens
                    )
                    logits = output.logits.view(-1, self.model.config.vocab_size)
            
                    pos_score = self.softmax(logits)[:, self.pos_id] # Yes
                    neg_score = self.softmax(logits)[:, self.neg_id] # No

                    cur_pos_score = [x.item() for x in pos_score]
                    cur_neg_score = [x.item() for x in neg_score]
                    pos_score_list += cur_pos_score
                    neg_score_list += cur_neg_score

            except RuntimeError:
                print(f'source: {src_list}')
                print(f'target: {tgt_list}')
                exit(0)
        
        score_list = []
        for i in range(len(pos_score_list)):
            score_list.append(pos_score_list[i] / (pos_score_list[i] + neg_score_list[i]))
            
        return score_list



'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

from .constraint_checker import evaluate_2step_prompt
from ..utils import example_format_checker

def constraint_scores(
        method = "openai/gpt-3.5-turbo", 
        data = [], 
        global_knowledge = ""
    ):

    data = example_format_checker(data)
    scores, meta_data = evaluate_2step_prompt(method, data, global_knowledge)
    return scores, meta_data



'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

from .prompts import constraint_identification_prompt, constraint_checking_prompt
from ..llm_wrapper import OpenAILLM, LocalLLM
from tqdm import tqdm


# Read all the possible inputs and decide in the evaluation function how to use them.
def format_model_inputs(sample):
    return sample["prompt_all"], sample["prompt_task"], sample["prompt_context"], sample["output"]

"""
main function that evaluates and scores how well the constraints in the prompt are satisfied by the output

TODO:
1. What score for no constraints? (full?) - current scores it as 1.0 if no constraints identified
2. How to use the local and global knowledge here? -- Initial thoughts: Local or Global knowledge is only useful for factual verification not for constraint check


Scoring the constraints in 2 steps: 
1. constraint identification using only prompt as the input
2. checking constraints as a set of rules against the LLM output
"""

def evaluate_2step_prompt(method, data, global_knowledge=''):

    constraint_scores = []
    score_reasoning = []

    if "openai" in method:
        model = OpenAILLM(model_name = method.split("/")[-1])
    else:
        model_task = "text-generation" if "gpt" in method else "text2text-generation"
        model =  LocalLLM(model_name = method, task = model_task)

    for sample in tqdm(data):
        prompt_all, task, input_doc, llm_output = format_model_inputs(sample)

        # populate the prompt
        if prompt_all:  # if instruction and document provided together, no option but to use both for identification
            prompt_identification = constraint_identification_prompt.format(instructions=prompt_all)
        else:
            prompt_identification = constraint_identification_prompt.format(instructions=task)

        # run the prompt 
        constraints_found = model.generate(prompt=prompt_identification, messages="")

        if 'No Constraints.' in constraints_found:
            constraint_scores.append(1.0)
            score_reasoning.append(constraints_found)
            continue

        # if specific constraints found
        prompt_checking = constraint_checking_prompt.format(llm_output=llm_output, constraints=constraints_found)

        constraints_checked = model.generate(prompt=prompt_checking, messages="")

        satisfied = []
        for constraint in constraints_checked.split("\n"):
            
            if constraint and 'Constraint' in constraint:

                # count the num of yes-es
                if "Yes" in constraint:
                    satisfied.append(1)
                else:
                    satisfied.append(0)
            
        score = sum(satisfied)/len(satisfied)
        
        # Use full response text as the reasoning
        constraint_scores.append(score)
        score_reasoning.append(constraints_found + "\n" + constraints_checked)

    # TODO: Add exact constraint checked back into the reasoning for it to make sense 
    meta_data = {
        "aspect_explanation": score_reasoning
    }

    # print ("\n==================\nconstraint_scores 2 step: ", constraint_scores)
    return constraint_scores, meta_data




from langchain import PromptTemplate
'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

gpt3_score_template = """{instruction}

Ensure that the generated text is factually accurate and consistent. Refer to local knowledge or general knowledge for missing or additional factual information.
Local Knowledge: {local_knowledge}
Global Knowledge: {global_knowledge}

Generated text: {output}"""

chatgpt_evaluator_template = """Score the output text, given the corresponding instruction, with respect to consistency using a scale of one to five stars. One star indicates "inconsistency" while five stars means "perfect consistency". Note that consistency measures whether the facts in the output are consistent with the facts in the instruction. Consider whether the output does reproduce all facts accurately and does not make up untrue information. The contextual text may include instruction and source text, local knowledge that is specific to the source text, and global knowledge.

Instruction: {instruction}
Local Knowledge: {local_knowledge}
Global Knowledge: {global_knowledge}
Output: {output}
Stars:"""

localllm_evaluator_template = """Is the Output factual given the corresponding instruction and knowledge? Give of score of 1, 2, 3 where 1 means seriously inconsistent, 2 means somewhat inconsistent, and 3 means consistent.

Instruction: {instruction}
Local Knowledge: {local_knowledge}
Global Knowledge: {global_knowledge}
Output: {output}

The score is:"""

opt_score_instruction_template = """{instruction}

Local Knowledge: {local_knowledge}
Global Knowledge: {global_knowledge}
"""

prompt_templates = {
    'chatgpt_evaluator': PromptTemplate(input_variables=["instruction", "local_knowledge", "global_knowledge", "output"], template=chatgpt_evaluator_template),
    'gpt3_score': PromptTemplate(input_variables=["instruction", "local_knowledge", "global_knowledge", "output"], template=gpt3_score_template),
    'localllm_score': PromptTemplate(input_variables=["instruction", "local_knowledge", "global_knowledge", "output"], template=localllm_evaluator_template),
    'opt_score': PromptTemplate(input_variables=["instruction", "local_knowledge", "global_knowledge"], template=opt_score_instruction_template),
}



import json, os, pandas as pd, numpy as np, csv
from datasets import load_dataset
from collections import Counter
import requests, zipfile, tarfile
import .utils_scorer import choose_best_threshold
import .utils_misc import download_file_from_google_drive

# SummaC Benchmark
class SummaCBenchmark:

    def __init__(self, benchmark_folder="/home/phillab/data/summac_benchmark/", dataset_names=["cogensum", "xsumfaith", "polytope", "factcc", "summeval", "frank"], cut="val"):
        assert cut in ["val", "test"], "Unrecognized cut for the Fact Checking Benchmark"
        if not os.path.exists(benchmark_folder):
            os.makedirs(benchmark_folder)

        self.cut = cut
        self.benchmark_folder = benchmark_folder
        self.cnndm_id2reference = None
        self.cnndm = None
        self.xsum = None

        self.datasets = []
        for dataset_name in dataset_names:
            if dataset_name == "cogensum":
                self.load_cogensumm()
            elif dataset_name == "xsumfaith":
                self.load_xsumfaith()
            elif dataset_name == "polytope":
                self.load_polytope()
            elif dataset_name == "factcc":
                self.load_factcc()
            elif dataset_name == "summeval":
                self.load_summeval()
            elif dataset_name == "frank":
                self.load_frank()
            else:
                raise ValueError("Unrecognized dataset name: %s" % (dataset_name))

    # Underlying dataset loader: CNN/DM and XSum
    def get_cnndm_document(self, aid):
        global CNNDM
        if self.cnndm is None:
            if CNNDM is None:
                CNNDM = load_dataset("cnn_dailymail", "3.0.0")
            self.cnndm = CNNDM
            self.cnndm_id2article = {}
            for cut in ["test", "validation"]:
                self.cnndm_id2article.update({d["id"]: d["article"] for d in self.cnndm[cut]})
        return self.cnndm_id2article[aid]

    def get_cnndm_reference(self, aid):
        global CNNDM
        if CNNDM is None:
            CNNDM = load_dataset("cnn_dailymail", "3.0.0")
            self.cnndm = CNNDM
        if self.cnndm_id2reference is None:
            self.cnndm_id2reference = {}
            for cut in ["test", "validation"]:
                self.cnndm_id2reference.update({d["id"]: d["highlights"] for d in self.cnndm[cut]})
        return self.cnndm_id2reference[aid]


    def get_xsum_document(self, aid):
        if self.xsum is None:
            self.xsum = load_dataset("xsum")["test"]
            self.xsumid2article = {d["id"]: d["document"] for d in self.xsum}

        return self.xsumid2article[aid]

    # Individual dataset loaders
    def load_cogensumm(self):
        # Correctness of Generated Summaries: https://www.aclweb.org/anthology/P19-1213.pdf
        # CoGenSumm: https://tudatalib.ulb.tu-darmstadt.de/handle/tudatalib/2002

        dataset_folder = os.path.join(self.benchmark_folder, "cogensumm/")
        if not os.path.exists(dataset_folder):
            print("==== CoGenSumm dataset not found, downloading from scratch")
            os.makedirs(dataset_folder)
            data = requests.get("https://tudatalib.ulb.tu-darmstadt.de/bitstream/handle/tudatalib/2002/summary-correctness-v1.0.zip?sequence=3&isAllowed=y")
            zip_file = os.path.join(dataset_folder, "summary-correctness-v1.0.zip")
            with open(zip_file, "wb") as f:
                f.write(data.content)

            with zipfile.ZipFile(zip_file, "r") as zip_ref:
                zip_ref.extractall(dataset_folder)
            os.remove(zip_file)

        clean_dataset = []
        for fn in os.listdir(dataset_folder):
            if self.cut not in fn:
                continue

            with open(os.path.join(dataset_folder, fn), "r") as f:
                dataset = json.load(f)

            if "_org" in fn or fn == "test_chen18_reranked.json":
                for aid in dataset:
                    document = self.get_cnndm_document(aid)
                    label = 0 if dataset[aid]["label"] == "Incorrect" else 1
                    sents = dataset[aid]["sents"]
                    summary = " ".join([sents[str(i)]["text"] for i in range(len(sents))])
                    clean_dataset.append({"filename": fn, "label": label, "document": document, "claim": summary, "cnndm_id": aid, "annotations": [label], "dataset": "cogensumm", "origin": "cnndm"})
            elif fn == "val_reranking.json":
                for aid in dataset:
                    document = self.get_cnndm_document(aid)
                    for idx, data in dataset[aid].items():
                        label = 0 if data["label"] == "Incorrect" else 1
                        summary = " ".join([data["sents"][str(i)]["text"] for i in range(len(data["sents"]))])
                        clean_dataset.append({"filename": fn, "label": label, "document": document, "claim": summary, "cnndm_id": aid, "annotations": [label], "dataset": "cogensumm", "origin": "cnndm"})
            elif fn == "val_sentence_pairs.json":
                for d in dataset:
                    aid = d["article_id"]
                    document = self.get_cnndm_document(aid)
                    clean_dataset.append({"filename": fn, "label": 1, "document": document, "claim": d["correct_sent"], "cnndm_id": aid, "annotations": [1], "dataset": "cogensumm", "origin": "cnndm"})
                    clean_dataset.append({"filename": fn, "label": 0, "document": document, "claim": d["incorrect_sent"], "cnndm_id": aid, "annotations": [0], "dataset": "cogensumm", "origin": "cnndm"})
        self.datasets.append({"name": "cogensumm", "dataset": clean_dataset})

    def load_xsumfaith(self):
        # On Faithfulness and Factuality in Abstractive Summarization - ACL 2020
        # https://github.com/google-research-datasets/xsum_hallucination_annotations
        # https://aclanthology.org/2020.acl-main.173.pdf

        dataset_folder = os.path.join(self.benchmark_folder, "xsumfaith/")
        if not os.path.exists(dataset_folder):
            print("==== XSum dataset not found, downloading from scratch")
            os.makedirs(dataset_folder)

            csv_file = requests.get("https://github.com/google-research-datasets/xsum_hallucination_annotations/raw/master/hallucination_annotations_xsum_summaries.csv")
            with open(os.path.join(dataset_folder, "hallucination_annotations_xsum_summaries.csv"), "wb") as f:
                f.write(csv_file.content)

        path_to_annotation = os.path.join(dataset_folder, "hallucination_annotations_xsum_summaries.csv")

        with open(path_to_annotation, "r") as f:
            raw_data = list(csv.reader(f))
            dataset = []
            keys = raw_data[0]
            for line in raw_data[1:]:
                dataset.append({k: v for k, v in zip(keys, line)})

        groups = {}
        for d in dataset:
            k = (d["bbcid"], d["system"])
            if k not in groups:
                groups[k] = []
            groups[k].append(d)

        clean_dataset = []
        for k, vs in groups.items():
            A = vs[0]
            document = self.get_xsum_document(A["bbcid"])
            labels = [v["hallucination_type"] for v in vs]
            annotations = [1 if label == "NULL" else 0 for label in labels]
            most_common_label = Counter(labels).most_common(1)[0][0]
            label = 1 if most_common_label == "NULL" else 0
            c = "val" if len(clean_dataset) % 2 == 0 else "test"

            clean_dataset.append({"document": document, "claim": A["summary"], "bbcid": A["bbcid"], "model_name": A["system"], "label": label, "cut": c, "annotations": annotations, "dataset": "xsumfaith", "origin": "xsum"})
        final_dataset = [d for d in clean_dataset if d["cut"]==self.cut]
        self.datasets.append({"name": "xsumfaith", "dataset": final_dataset})

    def load_polytope(self, which_label="overall"):
        # What Have We Achieved on Text Summarization? [https://arxiv.org/abs/2010.04529]
        # Dataset must be downloaded from the Github repo: https://github.com/hddbang/polytope

        assert which_label in ["overall", "omission", "addition", "duplication", "inaccuracy"], "Unrecognized `which label`"

        dataset_folder = os.path.join(self.benchmark_folder, "polytope")
        if not os.path.exists(dataset_folder):
            print("==== Polytope dataset not found, downloading from scratch")
            os.makedirs(dataset_folder)

            for model_name in ["BART", "Bert_Ext", "Bert_Ext_Abs", "BottomUp", "PG", "PG_Coverage", "Summa", "TextRank", "seq2seq"]:
                url = "https://github.com/hddbang/PolyTope/raw/master/outputs_with_human_annotation/Human_Annotation_Summarization_%s.xlsm" % (model_name)
                r = requests.get(url)
                with open(os.path.join(dataset_folder, "Human_Annotation_Summarization_%s.xlsm" % (model_name)), "wb") as f:
                    f.write(r.content)

        full_dataset = []
        for fn in os.listdir(dataset_folder):
            fn = os.path.join(dataset_folder, fn)

            all_segments = pd.read_excel(fn, sheet_name="Scores per segment")
            ID2row = {}
            for i, segment in all_segments.iterrows():
                c = "val" if i % 2 == 0 else "test"
                if str(segment["ID"]) != "nan":
                    ID2row[segment["ID"]] = {"ID": segment["ID"], "document": segment["Source"], "claim": segment["Target"], "errors": [], "cut": c}

            for i, row in pd.read_excel(fn, sheet_name="Error Log").iterrows():
                if str(row["Subtypes"]) != "nan":
                    ID2row[row["ID"]]["errors"].append(row["Subtypes"])

            for ID in ID2row:
                d = ID2row[ID]
                d["overall_label"] = 1 if len(d["errors"]) == 0 else 0
                d["omission_label"] = 0 if "Omission" in d["errors"] else 1
                d["addition_label"] = 0 if "Addition" in d["errors"] else 1
                d["duplication_label"] = 0 if "Duplication" in d["errors"] else 1
                d["inaccuracy_label"] = 0 if "Inaccuracy_internal" in d["errors"] or "Inaccuracy_external" in d["errors"] else 1
                if which_label is not None:
                    d["label"] = d["%s_label" % (which_label)]
                d["dataset"] = "polytope"
                d["annotations"] = [d["label"]]
                d["origin"] = "cnndm"

                full_dataset.append(d)
        cut_dataset = [d for d in full_dataset if d["cut"]==self.cut]
        self.datasets.append({"name": "polytope", "dataset": cut_dataset})

    def load_factcc(self, max_entries=-1):
        # Evaluating the Factual Consistency of Abstractive Text Summarization [https://arxiv.org/abs/1910.12840]
        # Dataset for each split must be downloaded from the Github repo: https://github.com/salesforce/factCC

        dataset_folder = os.path.join(self.benchmark_folder, "factcc/")
        if not os.path.exists(dataset_folder):
            print("==== FactCC dataset not found, downloading from scratch")
            os.makedirs(dataset_folder)

            urls = ["https://storage.googleapis.com/sfr-factcc-data-research/unpaired_generated_data.tar.gz", "https://storage.googleapis.com/sfr-factcc-data-research/unpaired_annotated_data.tar.gz"]
            for url in urls:
                zip_name = url.split("/")[-1]
                r = requests.get(url)
                with open(os.path.join(dataset_folder, zip_name), "wb") as f:
                    f.write(r.content)
                
                with tarfile.open(os.path.join(dataset_folder, zip_name), "r:gz") as f:
                    f.extractall(dataset_folder)
                os.remove(os.path.join(dataset_folder, zip_name))

        if self.cut == "train":
            dataset = []
            with open(os.path.join(dataset_folder, "unpaired_generated_data/data-original/data-train.jsonl"), "r") as f:
                for i, line in enumerate(f):
                    if max_entries > 0 and i >= max_entries:
                        break
                    D = json.loads(line)
                    aid = D["filepath"].split("/")[-1].replace(".story", "")
                    full_text = self.get_cnndm_document(aid)

                    label = 1 if D["label"]=="CORRECT" else 0
                    datum = {"document": full_text, "claim": D["claim"], "cnndm_id": D["id"], "label": label, "dataset": "factcc", "origin": "cnndm"}
                    dataset.append(datum)

        if self.cut in ["val", "test"]:
            factcc_file = os.path.join(dataset_folder, "unpaired_annotated_data/%s/data-dev.jsonl" % (self.cut))
            dataset = []
            with open(factcc_file, "r") as f:
                for line in f:
                    dataset.append(json.loads(line))

            for d in dataset:
                aid = d["filepath"].split("/")[-1].replace(".story", "")
                d["document"] = self.get_cnndm_document(aid)
                d["label"] = 1 if d["label"] == "CORRECT" else 0
                d["annotations"] = [d["label"]]
                d["dataset"] = "factcc"
                d["origin"] = "cnndm"

        self.datasets.append({"name": "factcc", "dataset": dataset})

    def load_summeval(self, key_focus="consistency"):
        assert key_focus in ["consistency", "coherence", "fluency", "relevance"]
        # SummEval: Re-evaluating Summarization Evaluation [https://arxiv.org/abs/2007.12626]
        # Data files must be downloaded from the following Github repository: https://github.com/Yale-LILY/SummEval
        raw_dataset = []

        dataset_folder = os.path.join(self.benchmark_folder, "summeval/")
        fn = os.path.join(dataset_folder, "model_annotations.aligned.scored.jsonl")
        if not os.path.exists(dataset_folder):
            print("==== SummEval dataset not found, downloading from scratch")
            os.makedirs(dataset_folder)

            # From the 4/19/2020 update on the README: https://github.com/Yale-LILY/SummEval
            download_file_from_google_drive("1d2Iaz3jNraURP1i7CfTqPIj8REZMJ3tS", fn)

        with open(fn, "r") as f:
            for line in f:
                raw_dataset.append(json.loads(line))

        clean_dataset = []

        for i, d in enumerate(raw_dataset):
            c = "val" if i % 2 == 0 else "test"
            _, _, article_id = d["id"].split("-")
            document = self.get_cnndm_document(article_id)
            annotations = d["expert_annotations"]

            consistencies = [a[key_focus] for a in annotations]
            final_label = 1 if len([cons for cons in consistencies if cons==5]) > len(annotations)/2 else 0

            # annotations = [1 if cons == 5 else 0 for cons in consistencies]
            annotations = consistencies
            error_type = "no error" if final_label == 1 else "error"

            clean_dataset.append({"document": document, "claim": d["decoded"], "label": final_label, "model_name": d["model_id"], "cnndm_id": d["id"], "cut": c, "annotations": annotations, "dataset": "summeval", "origin": "cnndm", "error_type": error_type})
        final_dataset = [d for d in clean_dataset if d["cut"] == self.cut]
        self.datasets.append({"name": "summeval", "dataset": final_dataset})

    def load_frank(self):
        # FRANK: Factuality Evaluation Benchmark [https://aclanthology.org/2021.naacl-main.383.pdf]
        # Files must be downloaded from the Github repository: https://github.com/artidoro/frank

        dataset_folder = os.path.join(self.benchmark_folder, "frank/")
        if not os.path.exists(dataset_folder):
            print("==== Frank dataset not found, downloading from scratch")
            os.makedirs(dataset_folder)

            fns = ["human_annotations_sentence.json", "validation_split.txt", "test_split.txt"]
            for fn in fns:
                data = requests.get("https://raw.githubusercontent.com/artidoro/frank/main/data/%s" % fn)
                with open(os.path.join(dataset_folder, fn), "w") as f:
                    f.write(data.text)

        raw_file = os.path.join(dataset_folder, "human_annotations_sentence.json")
        val_hash_file = os.path.join(dataset_folder, "validation_split.txt")
        test_hash_file = os.path.join(dataset_folder, "test_split.txt")
        with open(val_hash_file if self.cut=="val" else test_hash_file, "r") as f:
            valid_hashes = set([line.strip() for line in f])

        with open(raw_file, "r") as f:
            raw_dataset = json.load(f)
        dataset = []
        for d in raw_dataset:
            article = d["article"]
            origin = "cnndm" if len(d["hash"]) >= 40 else "xsum"

            if d["hash"] not in valid_hashes:
                continue

            summ_labels = []
            annotator_labels = {}
            for annot in d["summary_sentences_annotations"]:
                annot_vals = [an for ans in annot.values() for an in ans]
                noerror_count = len([an for an in annot_vals if an=="NoE"])
                label = 1 if noerror_count >= 2 else 0
                summ_labels.append(label)
                for anno_name, anno in annot.items():
                    if anno_name not in annotator_labels:
                        annotator_labels[anno_name] = []
                    annotator_labels[anno_name] += anno

            annotations = [1 if all(a=="NoE" for a in annos) else 0 for annos in annotator_labels.values()]
            label = 0 if any(sl==0 for sl in summ_labels) else 1

            error_type = "NoE"
            if label == 0:
                errors = [anno for annos in annotator_labels.values() for anno in annos if anno != "NoE"]
                error_type = Counter(errors).most_common(1)[0][0]

            summary = d["summary"]
            dataset.append({"document": article, "claim": summary, "label": label, "cut": self.cut, "hash": d["hash"], "model_name": d["model_name"], "annotations": annotations, "dataset": "frank", "origin": origin, "error_type": error_type})
        self.datasets.append({"name": "frank", "dataset": dataset})

    def get_dataset(self, dataset_name):
        for dataset in self.datasets:
            if dataset["name"] == dataset_name:
                return dataset["dataset"]
        raise ValueError("Unrecognized dataset name: %s" % (dataset_name))

    def print_stats(self):
        dataset_stats = []
        for dataset in self.datasets:
            N_pos, N_neg = len([d for d in dataset["dataset"] if d["label"]==1]), len([d for d in dataset["dataset"] if d["label"]==0])
            dataset_stats.append({"name": dataset["name"], "N": len(dataset["dataset"]), "N_pos": N_pos, "N_neg": N_neg, "frac_pos": N_pos/(N_pos+N_neg)})
        print(pd.DataFrame(dataset_stats))

    def evaluate(self, scorer):
        benchmark = []

        for dataset in self.datasets:
            dataset_labels = [d["label"] for d in dataset["dataset"]]
            dataset_preds = scorer.score([d["document"] for d in dataset["dataset"]], [d["claim"] for d in dataset["dataset"]])["scores"]

            dataset_thresh, dataset_f1 = choose_best_threshold(dataset_labels, dataset_preds)
            benchmark.append({"name": dataset["name"], "score": dataset_f1, "threshold": dataset_thresh})
        return {"overall_score": np.mean([t["score"] for t in benchmark]), "benchmark": benchmark}


if __name__ == "__main__":
    import random

    for cut in ["val", "test"]:
        summac_benchmark = SummaCBenchmark(benchmark_folder="/home/tingu/data/summac_benchmark2/", cut=cut)
        print("============= SUMMAC %s ===============" % (cut.upper()))
        summac_benchmark.print_stats()
        for dataset in summac_benchmark.datasets:
            print("\n============= %s ===============" % (dataset["name"]))
            random.shuffle(dataset["dataset"])
            print(dataset["dataset"][0]["document"][:400])
            print("-------------")
            print(dataset["dataset"][0]["claim"])
            print("-------------")
            print(dataset["dataset"][0]["label"])



import datasets, nltk, numpy as np
import json, os, sys, argparse

class BaselineScorer:
    def __init__(self, model="questeval", do_weighter=False, load_cache=True):
        assert model in ["questeval", "feqa", "dae"], "Unrecognized baseline model"
        self.model = model
        self.do_weighter = do_weighter
        self.model_loaded = False
        self.cache = {}
        self.cache_file = "/export/share/plaban/summac_cache/cache_%s.json" % (self.model)
        if load_cache:
            self.load_cache()

    def load_model(self):
        if self.model == "questeval":
            from questeval.questeval_metric import QuestEval
            self.questeval = QuestEval(isCuda=True, do_weighter=self.do_weighter)
        elif self.model == "feqa":
            # import benepar, nltk
            # benepar.download('benepar_en2')
            # nltk.download('stopwords')
            from feqa import FEQA
            self.scorer = FEQA(use_gpu=True)
        elif self.model == "dae":
            sys.path.insert(0, "/home/phillab/dae-factuality/")
            from evaluate_factuality import MODEL_CLASSES, score_example_single_context

            parser = argparse.ArgumentParser()
            args = parser.parse_args()
            args.device = "cuda:0"
            args.per_gpu_eval_batch_size = 8
            args.max_seq_length = 128
            args.dependency_type =  "enhancedDependencies"
            self.args = args

            model_dir = "/home/phillab/models/dae_basic/"
            model_type = "electra_dae"
            config_class, model_class, tokenizer_class = MODEL_CLASSES[model_type]

            self.tokenizer = tokenizer_class.from_pretrained(model_dir)
            self.dae_model = model_class.from_pretrained(model_dir)
            self.dae_model.to(args.device)

        self.model_loaded = True

    def load_cache(self):
        if os.path.isfile(self.cache_file):
            with open(self.cache_file, "r") as f:
                self.cache = json.load(f)

    def save_cache(self):
        with open(self.cache_file, "w") as f:
            json.dump(self.cache, f)

    def get_sample_key(self, document, generated):
        return "%s|%%%%|%%|%s" % (document, generated)

    def score_questeval(self, documents, generateds, **kwargs):
        scores = []
        for document, generated in zip(documents, generateds):
            score = self.questeval.compute_all(document, generated)
            scores.append(score["scores"]["fscore"])
        return {"scores": scores}

    def score_feqa(self, documents, generateds, **kwargs):
        scores = self.scorer.compute_score(documents, generateds, aggregate=False)
        self.save_cache()
        return {"scores": scores}

    def score_dae(self, documents, generateds, **kwargs):
        from evaluate_factuality import score_example_single_context

        scores = []
        for document, generated in zip(documents, generateds):
            document = " ".join(document.split(" ")[:250])
            score = score_example_single_context(generated, document, self.dae_model, self.tokenizer, self.args).item()
            scores.append(score)

        # self.save_cache()
        return {"scores": scores}

    def score(self, documents, generateds, **kwargs):
        new_samples = []
        for d, g in zip(documents, generateds):
            k = self.get_sample_key(d, g)
            if k not in self.cache:
                new_samples.append((k, d, g))

        if len(new_samples) > 0:
            if not self.model_loaded:
                self.load_model()

            if self.model == "questeval":
                new_scores = self.score_questeval([d[1] for d in new_samples], [d[2] for d in new_samples])
            elif self.model == "feqa":
                new_scores = self.score_feqa([d[1] for d in new_samples], [d[2] for d in new_samples])
            elif self.model == "dae":
                new_scores = self.score_dae([d[1] for d in new_samples], [d[2] for d in new_samples])

            for (k, d, g), score in zip(new_samples, new_scores["scores"]):
                self.cache[k] = score

        return {"scores": [self.cache[self.get_sample_key(d, g)] for d, g in zip(documents, generateds)]}


if __name__ == "__main__":
    hypothesis = """After wildfires consumed an entire town, students and teachers who had planned for remote classes found some comfort in staying connected amid the chaos."""

    source = """Ash fell from an apocalyptic orange sky as Jennifer Willin drove home last week from the only school in tiny Berry Creek, Calif., where she had picked up a pair of Wi-Fi hot spots for her daughters’ remote classes. Hours later,
    her cellphone erupted with an emergency alert: Evacuate immediately. By the next morning, what one official described as a “massive wall of fire” had swept through the entire Northern California town of about 1,200 people, killing nine residents,
    including a 16-year-old boy, and destroying the school and almost every home and business. Ms. Willin and her family escaped to a cramped hotel room 60 miles away.
    In her panic, she had forgotten to grab masks, but she had the hot spots, along with her daughters’ laptops and school books. On Monday, the two girls plan to meet with their teachers on Zoom, seeking some comfort amid the chaos.
    They’re still able to be in school,” Ms. Willin said, “even though the school burned to the ground.”
    As the worst wildfire season in decades scorches the West amid a still raging pandemic, families and educators who were already starting the strangest and most challenging school year of their lifetimes have been traumatized all over again.
    Tens of thousands of people have been forced to flee their homes, with some mourning the loss of their entire communities.
    But amid the twin disasters, the remote learning preparations that schools made for the coronavirus are providing a strange
    modicum of stability for teachers and students, letting many stay connected and take comfort in an unexpected form of virtual community."""

    qe_score = BaselineScorer(model="dae")
    print(qe_score.score([source], [hypothesis]))



# borrow code from https://github.com/timoschick/self-debiasing

import itertools
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

import torch
from torch.nn import CrossEntropyLoss
from transformers import T5Tokenizer, T5ForConditionalGeneration, GPT2Tokenizer, PreTrainedTokenizer, PreTrainedModel

from .generation import SelfDebiasingGPT2LMHeadModel


class ModelWrapper(ABC):
    """
    This class represents a wrapper for a pretrained language model that provides some high-level functions, including zero-shot
    classification using cloze questions and the generation of texts with self-debiasing.
    """

    def __init__(self, use_cuda: bool = True):
        """
        :param use_cuda: whether to use CUDA
        """
        self._device = "cuda" if torch.cuda.is_available() and use_cuda else "cpu"
        self._tokenizer = None  # type: Optional[PreTrainedTokenizer]
        self._model = None  # type: Optional[PreTrainedModel]

    def query_model(self, input_text: str) -> torch.FloatTensor:
        """For a given input text, returns the probability distribution over possible next tokens."""
        return self.query_model_batch([input_text])[0]

    @abstractmethod
    def query_model_batch(self, input_texts: List[str]) -> torch.FloatTensor:
        """For a batch of input texts, returns the probability distribution over possible next tokens."""
        pass

    @abstractmethod
    def generate(self, input_text: str, **kwargs) -> str:
        """Generates a continuation for a given input text."""
        pass

    @abstractmethod
    def generate_self_debiasing(self, input_texts: List[str], debiasing_prefixes: List[str], decay_constant: float = 50,
                                epsilon: float = 0.01, debug: bool = False, **kwargs) -> List[str]:
        """
        Generates continuations for the given input texts with self-debiasing.
        :param input_texts: the input texts to generate continuations for
        :param debiasing_prefixes: the debiasing prefixes to be used
        :param decay_constant: the decay constant (lambda in the paper)
        :param epsilon: the minimum factor by which each probability is multiplied
        :param debug: whether to print additional debugging output
        :param kwargs: further arguments are passed on to the original generate function
        :return: the list of generated continuations
        """
        pass

    @abstractmethod
    def compute_loss(self, input_ids: torch.LongTensor, labels: torch.LongTensor) -> torch.Tensor:
        """Computes cross-entropy loss for the given input ids and corresponding labels."""
        pass

    @abstractmethod
    def compute_loss_self_debiasing(self, input_ids: torch.Tensor, trg_len: int, debiasing_prefixes: List[str], decay_constant: float = 50,
                                    epsilon: float = 0.01, debug: bool = False) -> torch.Tensor:
        """
        Computes cross-entropy loss for the given input ids with self-debiasing.
        :param input_ids: the input ids
        :param trg_len: only the last trg_len tokens are considered for computing the loss
        :param debiasing_prefixes: the debiasing prefixes to be used
        :param decay_constant: the decay constant (lambda in the paper)
        :param epsilon: the minimum factor by which each probability is multiplied
        :param debug: whether to print additional debugging output
        :return: the cross entropy loss
        """
        pass

    def get_token_probability_distribution(self, input_texts: List[str], output_choices: List[str]) -> List[List[Tuple[str, float]]]:
        """
        For a batch of input texts, returns the probability distribution over possible next tokens considering only the given list of
        output choices.
        :param input_texts: the input texts
        :param output_choices: the allowed output choices (must correspond to single tokens in the model's vocabulary)
        :return: a list of lists, where output[i][j] is a (output, probability) tuple for the ith input and jth output choice.
        """
        output_choice_ids = []
        kwargs = {'add_prefix_space': True} if isinstance(self, GPT2Wrapper) else {}
        for word in output_choices:
            tokens = self._tokenizer.tokenize(word, **kwargs)
            assert len(tokens) == 1, f"Word {word} consists of multiple tokens: {tokens}"
            assert tokens[0] not in self._tokenizer.all_special_tokens, f"Word {word} corresponds to a special token: {tokens[0]}"
            token_id = self._tokenizer.convert_tokens_to_ids(tokens)[0]
            output_choice_ids.append(token_id)

        logits = self.query_model_batch(input_texts)
        result = []

        for idx, _ in enumerate(input_texts):
            output_probabilities = logits[idx][output_choice_ids].softmax(dim=0)
            choices_with_probabilities = list(zip(output_choices, (prob.item() for prob in output_probabilities)))
            result.append(choices_with_probabilities)

        return result


class T5Wrapper(ModelWrapper):
    """A wrapper for the T5 model"""

    def __init__(self, model_name: str = "google/t5-v1_1-xl", use_cuda: bool = True):
        """
        :param model_name: the name of the pretrained T5 model (default: "google/t5-v1_1-xl")
        :param use_cuda: whether to use CUDA
        """
        super().__init__(use_cuda=use_cuda)
        self._tokenizer = T5Tokenizer.from_pretrained(model_name)
        self._model = T5ForConditionalGeneration.from_pretrained(model_name)
        if use_cuda:
            self._model.parallelize()

    def query_model_batch(self, input_texts: List[str]):
        assert all('<extra_id_0>' in input_text for input_text in input_texts)
        output_texts = ['<extra_id_0>'] * len(input_texts)
        inputs = self._tokenizer.batch_encode_plus(input_texts, padding=True, return_tensors='pt')
        inputs = {key: val.to(self._device) for key, val in inputs.items()}
        output_ids = self._tokenizer.batch_encode_plus(output_texts, return_tensors='pt')['input_ids'].to(self._device)
        return self._model(labels=output_ids, **inputs)['logits'][:, 1, :]

    def generate(self, input_text: str, **kwargs):
        assert '<extra_id_0>' in input_text
        input_ids = self._tokenizer.encode(input_text, return_tensors='pt').to(self._device)
        output_ids = self._model.generate(input_ids, **kwargs)[0]
        return self._tokenizer.decode(output_ids)

    def generate_self_debiasing(self, input_texts: List[str], debiasing_prefixes: List[str], decay_constant: float = 50,
                                epsilon: float = 0.01, debug: bool = False, **kwargs) -> List[str]:
        raise NotImplementedError()

    def compute_loss(self, input_ids: torch.LongTensor, labels: torch.LongTensor) -> torch.Tensor:
        raise NotImplementedError()

    def compute_loss_self_debiasing(self, input_ids: torch.Tensor, trg_len: int, debiasing_prefixes: List[str], decay_constant: float = 50,
                                    epsilon: float = 0.01, debug: bool = False) -> torch.Tensor:
        raise NotImplementedError()


class GPT2Wrapper(ModelWrapper):

    def __init__(self, model_name: str = "gpt2-xl", use_cuda: bool = True):
        """
        :param model_name: the name of the pretrained GPT2 model (default: "gpt2-xl")
        :param use_cuda: whether to use CUDA
        """
        super().__init__(use_cuda=use_cuda)
        self._tokenizer = GPT2Tokenizer.from_pretrained(model_name)
        self._model = SelfDebiasingGPT2LMHeadModel.from_pretrained(model_name)  # type: SelfDebiasingGPT2LMHeadModel
        if use_cuda:
            self._model.parallelize()
        self._tokenizer.pad_token = self._tokenizer.eos_token
        self._model.config.pad_token_id = self._tokenizer.eos_token_id

    def query_model_batch(self, input_texts: List[str]):
        inputs = self._tokenizer.batch_encode_plus(input_texts, padding=True, return_tensors='pt', truncation=True)
        inputs = {key: val.to(self._device) for key, val in inputs.items()}
        output_indices = inputs['attention_mask'].sum(dim=1) - 1
        output = self._model(**inputs)['logits']
        return torch.stack([output[example_idx, last_word_idx, :] for example_idx, last_word_idx in enumerate(output_indices)])

    def generate(self, input_text: str, **kwargs):
        input_ids = self._tokenizer.encode(input_text, return_tensors='pt').to(self._device)
        output_ids = self._model.generate(input_ids, **kwargs)[0]
        return self._tokenizer.decode(output_ids)

    def generate_self_debiasing(self, input_texts: List[str], debiasing_prefixes: List[str], decay_constant: float = 50,
                                epsilon: float = 0.01, debug: bool = False, min_length: int = None, max_length: int = None,
                                **kwargs) -> List[str]:

        self._model.init_logits_processor(num_debiasing_prefixes=len(debiasing_prefixes), decay_constant=decay_constant, epsilon=epsilon,
                                          debug=debug, tokenizer=self._tokenizer)
        inputs = input_texts.copy()
        for debiasing_prefix in debiasing_prefixes:
            for input_text in input_texts:
                inputs += [debiasing_prefix + input_text]

        inputs = self._tokenizer.batch_encode_plus(inputs, padding=True, return_tensors='pt')
        inputs['attention_mask'] = torch.flip(inputs['attention_mask'], dims=[1])
        shifts = inputs['attention_mask'].shape[-1] - inputs['attention_mask'].sum(dim=-1)
        for batch_idx in range(inputs['input_ids'].shape[0]):
            inputs['input_ids'][batch_idx] = inputs['input_ids'][batch_idx].roll(shifts[batch_idx].item())

        inputs = {k: v.to(self._device) for k, v in inputs.items()}
        input_length = inputs['input_ids'].shape[1]
        if min_length is not None:
            min_length = min_length + input_length
        if max_length is not None:
            max_length = max_length + input_length

        output_ids = self._model.generate(**inputs, min_length=min_length, max_length=max_length, **kwargs)

        batch_size = output_ids.shape[0] // (1 + len(debiasing_prefixes))
        output_ids = output_ids[:batch_size, inputs['input_ids'].shape[1]:]
        return self._tokenizer.batch_decode(output_ids)

    def compute_loss(self, input_ids: torch.LongTensor, labels: torch.LongTensor) -> torch.Tensor:
        outputs = self._model(input_ids, labels=labels)
        lm_logits = outputs[1]

        # Shift so that tokens < n predict n
        shift_logits = lm_logits[..., :-1, :].contiguous()
        shift_labels = labels[..., 1:].contiguous()
        # Flatten the tokens
        loss_fct = CrossEntropyLoss()
        loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))
        return loss

    def compute_loss_self_debiasing(self, input_ids: torch.Tensor, trg_len: int, debiasing_prefixes: List[str], decay_constant: float = 50,
                                    epsilon: float = 0.01, debug: bool = False) -> torch.Tensor:

        self._model.init_logits_processor(num_debiasing_prefixes=len(debiasing_prefixes), decay_constant=decay_constant, epsilon=epsilon,
                                          debug=debug, tokenizer=self._tokenizer)

        input_prefixes = [''] + debiasing_prefixes
        input_prefixes = self._tokenizer.batch_encode_plus(input_prefixes, padding=True, return_tensors='pt')
        input_prefixes['attention_mask'] = torch.flip(input_prefixes['attention_mask'], dims=[1])

        shifts = input_prefixes['attention_mask'].shape[-1] - input_prefixes['attention_mask'].sum(dim=-1)
        for batch_idx in range(input_prefixes['input_ids'].shape[0]):
            input_prefixes['input_ids'][batch_idx] = input_prefixes['input_ids'][batch_idx].roll(shifts[batch_idx].item())

        input_prefixes = {k: v.to(self._device) for k, v in input_prefixes.items()}

        input_ids_repeated = input_ids.repeat(len(debiasing_prefixes) + 1, 1)
        attention_mask = torch.ones_like(input_ids_repeated)

        attention_mask = torch.cat([input_prefixes['attention_mask'], attention_mask], dim=-1)
        input_ids_repeated = torch.cat([input_prefixes['input_ids'], input_ids_repeated], dim=-1)

        target_ids = input_ids_repeated.clone()
        trg_len += shifts[0]
        target_ids[:, :-trg_len] = -100

        position_ids = attention_mask.long().cumsum(-1) - 1
        position_ids.masked_fill_(attention_mask == 0, 1)

        outputs = self._model(input_ids=input_ids_repeated, attention_mask=attention_mask, position_ids=position_ids, labels=target_ids)
        lm_logits = outputs[1]

        for idx in range(lm_logits.shape[1]):
            lm_logits[:, idx, :] = self._model.logits_processor(input_ids=None, scores=lm_logits[:, idx, :])

        batch_size = lm_logits.shape[0] // (1 + len(debiasing_prefixes))
        lm_logits = lm_logits[:batch_size, shifts[0]:, :]
        target_ids = target_ids[:batch_size, shifts[0]:]

        # Shift so that tokens < n predict n
        shift_logits = lm_logits[..., :-1, :].contiguous()
        shift_labels = target_ids[..., 1:].contiguous()
        # Flatten the tokens
        loss_fct = CrossEntropyLoss()
        loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))
        return loss


'''
 * Copyright (c) 2023, Salesforce, Inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
'''

from .summac.summac.model_summac import SummaCConv
from ..utils import retrieve_global_knowledge, get_instruction

class SummacEvaluator:
    def __init__(self, device='cuda:0'):
        """ Set up evaluator for factual consistency detection """
        self.model_conv = SummaCConv(models=["vitc"], bins='percentile', granularity="sentence", \
            nli_labels="e", device=device, start_file="default", agg="mean")

    def evaluate(self, data, global_knowledge):
        import nltk
        nltk.download('punkt')
        sources = []
        generateds = []
        for sample in data:
            source = '\n'.join([get_instruction(sample),
                                sample['knowledge'],
                                retrieve_global_knowledge(sample['output'], global_knowledge)])
            sources.append(source)
            generateds.append(sample['output'])
        scores = self.model_conv.score(sources, generateds)["scores"]
        return scores



from transformers import AutoTokenizer, AutoModelForSequenceClassification
import nltk, numpy as np, torch, os, json
from .utils_misc import batcher

model_map = {
    "snli-base": {"model_card": "boychaboy/SNLI_roberta-base", "entailment_idx": 0, "contradiction_idx": 2},
    "snli-large": {"model_card": "boychaboy/SNLI_roberta-large", "entailment_idx": 0, "contradiction_idx": 2},
    "mnli-base": {"model_card": "microsoft/deberta-base-mnli", "entailment_idx": 2, "contradiction_idx": 0},
    "mnli": {"model_card": "roberta-large-mnli", "entailment_idx": 2, "contradiction_idx": 0},
    "anli": {"model_card": "ynie/roberta-large-snli_mnli_fever_anli_R1_R2_R3-nli", "entailment_idx": 0, "contradiction_idx": 2},
    "vitc-base": {"model_card": "tals/albert-base-vitaminc-mnli", "entailment_idx": 0, "contradiction_idx": 1},
    "vitc": {"model_card": "tals/albert-xlarge-vitaminc-mnli", "entailment_idx": 0, "contradiction_idx": 1},
    "vitc-only": {"model_card": "tals/albert-xlarge-vitaminc", "entailment_idx": 0, "contradiction_idx": 1},
    # "decomp": 0,
}

def card_to_name(card):
    card2name = {v["model_card"]: k for k, v in model_map.items()}
    if card in card2name:
        return card2name[card]
    return card

def name_to_card(name):
    if name in model_map:
        return model_map[name]["model_card"]
    return name

def get_neutral_idx(ent_idx, con_idx):
    return list(set([0, 1, 2]) - set([ent_idx, con_idx]))[0]

class SummaCImager:
    def __init__(self, model_name="mnli", granularity="paragraph", use_cache=True, max_doc_sents=100, device="cuda", **kwargs):

        self.grans = granularity.split("-")

        assert all(gran in ["paragraph", "sentence", "document", "2sents", "mixed"] for gran in self.grans) and len(self.grans) <= 2, "Unrecognized `granularity` %s" % (granularity)
        assert model_name in model_map.keys(), "Unrecognized model name: `%s`" % (model_name)

        self.model_name = model_name
        if model_name != "decomp":
            self.model_card = name_to_card(model_name)
            self.entailment_idx = model_map[model_name]["entailment_idx"]
            self.contradiction_idx = model_map[model_name]["contradiction_idx"]
            self.neutral_idx = get_neutral_idx(self.entailment_idx, self.contradiction_idx)

        self.granularity = granularity
        self.use_cache = use_cache
        self.cache_folder = "/export/share/plaban/summac_cache/"

        self.max_doc_sents = max_doc_sents
        self.max_input_length = 500
        self.device = device
        self.cache = {}
        self.model = None # Lazy loader

    def load_nli(self):
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_card)
        self.model = AutoModelForSequenceClassification.from_pretrained(self.model_card).eval()
        self.model.to(self.device)
        if self.device == "cuda":
            self.model.half()

    def split_sentences(self, text):
        sentences = nltk.tokenize.sent_tokenize(text)
        sentences = [sent for sent in sentences if len(sent)>10]
        return sentences

    def split_2sents(self, text):
        sentences = nltk.tokenize.sent_tokenize(text)
        sentences = [sent for sent in sentences if len(sent)>10]
        two_sents = [" ".join(sentences[i:(i+2)]) for i in range(len(sentences))]
        return two_sents

    def split_paragraphs(self, text):
        if text.count("\n\n") > 0:
            paragraphs = [p.strip() for p in text.split("\n\n")]
        else:
            paragraphs = [p.strip() for p in text.split("\n")]
        return [p for p in paragraphs if len(p) > 10]

    def split_text(self, text, granularity="sentence"):
        if granularity == "document":
            return [text]
        elif granularity == "paragraph":
            return self.split_paragraphs(text)
        elif granularity == "sentence":
            return self.split_sentences(text)
        elif granularity == "2sents":
            return self.split_2sents(text)
        elif granularity == "mixed":
            return self.split_sentences(text) + self.split_paragraphs(text)

    def build_chunk_dataset(self, original, generated, pair_idx=None):
        if len(self.grans) == 1:
            gran_doc, gran_sum = self.grans[0], self.grans[0]
        else:
            gran_doc, gran_sum = self.grans[0], self.grans[1]

        original_chunks = self.split_text(original, granularity=gran_doc)[:self.max_doc_sents]
        generated_chunks = self.split_text(generated, granularity=gran_sum)

        N_ori, N_gen = len(original_chunks), len(generated_chunks)
        dataset = [{"premise": original_chunks[i], "hypothesis": generated_chunks[j], "doc_i": i, "gen_i": j, "pair_idx": pair_idx} for i in range(N_ori) for j in range(N_gen)]
        return dataset, N_ori, N_gen

    def build_image(self, original, generated):
        cache_key = (original, generated)
        if self.use_cache and cache_key in self.cache:
            cached_image = self.cache[cache_key]
            cached_image = cached_image[:, :self.max_doc_sents, :]
            return cached_image

        dataset, N_ori, N_gen = self.build_chunk_dataset(original, generated)
        
        if len(dataset) == 0:
            return np.zeros((3, 1, 1))

        image = np.zeros((3, N_ori, N_gen))

        if self.model is None:
            self.load_nli()

        for batch in batcher(dataset, batch_size=20):
            batch_prems = [b["premise"] for b in batch]
            batch_hypos = [b["hypothesis"] for b in batch]
            batch_tokens = self.tokenizer.batch_encode_plus(list(zip(batch_prems, batch_hypos)), padding=True, truncation=True, max_length=self.max_input_length, return_tensors="pt", truncation_strategy="only_first")
            with torch.no_grad():
                model_outputs = self.model(**{k: v.to(self.device) for k, v in batch_tokens.items()})

            batch_probs = torch.nn.functional.softmax(model_outputs["logits"], dim=-1)
            batch_evids = batch_probs[:, self.entailment_idx].tolist()
            batch_conts = batch_probs[:, self.contradiction_idx].tolist()
            batch_neuts = batch_probs[:, self.neutral_idx].tolist()

            for b, evid, cont, neut in zip(batch, batch_evids, batch_conts, batch_neuts):
                image[0, b["doc_i"], b["gen_i"]] = evid
                image[1, b["doc_i"], b["gen_i"]] = cont
                image[2, b["doc_i"], b["gen_i"]] = neut

        if self.use_cache:
            self.cache[cache_key] = image
        return image

    def build_images(self, originals, generateds, batch_size=128):
        todo_originals, todo_generateds = [], []
        for ori, gen in zip(originals, generateds):
            cache_key = (ori, gen)
            if cache_key not in self.cache:
                todo_originals.append(ori)
                todo_generateds.append(gen)
        
        total_dataset = []
        todo_images = []
        for pair_idx, (ori, gen) in enumerate(zip(todo_originals, todo_generateds)):
            dataset, N_ori, N_gen = self.build_chunk_dataset(ori, gen, pair_idx=pair_idx)
            if len(dataset) == 0:
                image = np.zeros((3, 1, 1))
            else:
                image = np.zeros((3, N_ori, N_gen))
            todo_images.append(image)
            total_dataset += dataset
        if len(total_dataset) > 0 and self.model is None: # Can't just rely on the cache
            self.load_nli()
        
        for batch in batcher(total_dataset, batch_size=batch_size):
            batch_prems = [b["premise"] for b in batch]
            batch_hypos = [b["hypothesis"] for b in batch]
            batch_tokens = self.tokenizer.batch_encode_plus(list(zip(batch_prems, batch_hypos)), padding=True, truncation=True, max_length=self.max_input_length, return_tensors="pt", truncation_strategy="only_first")
            with torch.no_grad():
                model_outputs = self.model(**{k: v.to(self.device) for k, v in batch_tokens.items()})

            batch_probs = torch.nn.functional.softmax(model_outputs["logits"], dim=-1)
            batch_evids = batch_probs[:, self.entailment_idx].tolist()
            batch_conts = batch_probs[:, self.contradiction_idx].tolist()
            batch_neuts = batch_probs[:, self.neutral_idx].tolist()

            for b, evid, cont, neut in zip(batch, batch_evids, batch_conts, batch_neuts):
                image = todo_images[b["pair_idx"]]
                image[0, b["doc_i"], b["gen_i"]] = evid
                image[1, b["doc_i"], b["gen_i"]] = cont
                image[2, b["doc_i"], b["gen_i"]] = neut

        for pair_idx, (ori, gen) in enumerate(zip(todo_originals, todo_generateds)):
            cache_key = (ori, gen)
            self.cache[cache_key] = todo_images[pair_idx]
        
        images = [self.cache[(ori, gen)] for ori, gen in zip(originals, generateds)]
        return images


    def get_cache_file(self):
        return os.path.join(self.cache_folder, "cache_%s_%s.json" % (self.model_name, self.granularity))

    def save_cache(self):
        cache_cp = {"[///]".join(k): v.tolist() for k, v in self.cache.items()}
        with open(self.get_cache_file(), "w") as f:
            json.dump(cache_cp, f)

    def load_cache(self):
        cache_file = self.get_cache_file()
        if os.path.isfile(cache_file):
            with open(cache_file, "r") as f:
                cache_cp = json.load(f)
                self.cache = {tuple(k.split("[///]")): np.array(v) for k, v in cache_cp.items()}

class SummaCConv(torch.nn.Module):
    def __init__(self, models=["mnli", "anli", "vitc"], bins='even50', granularity="sentence", nli_labels="e", device="cuda", start_file=None, imager_load_cache=True, agg="mean", **kwargs):
        # `bins` should be `even%d` or `percentiles`
        assert nli_labels in ["e", "c", "n", "ec", "en", "cn", "ecn"], "Unrecognized nli_labels argument %s" % (nli_labels)

        super(SummaCConv, self).__init__()
        self.device = device
        self.models = models

        self.imagers = []
        for model_name in models:
            self.imagers.append(SummaCImager(model_name=model_name, granularity=granularity, device=self.device, **kwargs))
        if imager_load_cache:
            for imager in self.imagers:
                imager.load_cache()
        assert len(self.imagers)>0, "Imager names were empty or unrecognized"

        if "even" in bins:
            n_bins = int(bins.replace("even", ""))
            self.bins = list(np.arange(0, 1, 1/n_bins)) + [1.0]
        elif bins == "percentile":
            self.bins = [0.0, 0.01, 0.02, 0.03, 0.04, 0.07, 0.13, 0.37, 0.90, 0.91, 0.92, 0.93, 0.94, 0.95, 0.955, 0.96, 0.965, 0.97, 0.975, 0.98, 0.985, 0.99, 0.995, 1.0] # Based on the percentile of the distribution on some large number of summaries

        self.nli_labels = nli_labels
        self.n_bins = len(self.bins) - 1
        self.n_rows = 10
        self.n_labels = 2
        self.n_depth = len(self.imagers)*len(self.nli_labels)
        self.full_size = self.n_depth*self.n_bins

        self.agg = agg

        self.mlp = torch.nn.Linear(self.full_size, 1).to(device)
        self.layer_final = torch.nn.Linear(3, self.n_labels).to(device)

        if start_file == "default":
            start_file = "summac_conv_vitc_sent_perc_e.bin"
            if not os.path.isfile("summac_conv_vitc_sent_perc_e.bin"):
                os.system("wget https://github.com/tingofurro/summac/raw/master/summac_conv_vitc_sent_perc_e.bin")
                assert bins == "percentile", "bins mode should be set to percentile if using the default 1-d convolution weights."
        if start_file is not None:
            print(self.load_state_dict(torch.load(start_file)))

    def build_image(self, original, generated):
        images = [imager.build_image(original, generated) for imager in self.imagers]
        image = np.concatenate(images, axis=0)
        return image

    def compute_histogram(self, original=None, generated=None, image=None):
        # Takes the two texts, and generates a (n_rows, 2*n_bins)

        if image is None:
            image = self.build_image(original, generated)

        N_depth, N_ori, N_gen = image.shape

        full_histogram = []
        for i_gen in range(N_gen):
            histos = []

            for i_depth in range(N_depth):
                if (i_depth % 3 == 0 and "e" in self.nli_labels) or (i_depth % 3 == 1 and "c" in self.nli_labels) or (i_depth % 3 == 2 and "n" in self.nli_labels):
                    histo, X = np.histogram(image[i_depth, :, i_gen], range=(0, 1), bins=self.bins, density=False)
                    histos.append(histo)

            histogram_row = np.concatenate(histos)
            full_histogram.append(histogram_row)

        n_rows_missing = self.n_rows - len(full_histogram)
        full_histogram += [[0.0] * self.full_size] * n_rows_missing
        full_histogram = full_histogram[:self.n_rows]
        full_histogram = np.array(full_histogram)
        return image, full_histogram

    def forward(self, originals, generateds, images=None):
        if images is not None:
            # In case they've been pre-computed.
            histograms = []
            for image in images:
                _, histogram = self.compute_histogram(image=image)
                histograms.append(histogram)
        else:
            images, histograms = [], []
            for original, generated in zip(originals, generateds):
                image, histogram = self.compute_histogram(original=original, generated=generated)
                images.append(image)
                histograms.append(histogram)

        N = len(histograms)
        histograms = torch.FloatTensor(histograms).to(self.device)

        non_zeros = (torch.sum(histograms, dim=-1) != 0.0).long()
        seq_lengths = non_zeros.sum(dim=-1).tolist()

        mlp_outs = self.mlp(histograms).reshape(N, self.n_rows)
        features = []

        for mlp_out, seq_length in zip(mlp_outs, seq_lengths):
            if seq_length > 0:
                Rs = mlp_out[:seq_length]
                if self.agg == "mean":
                    features.append(torch.cat([torch.mean(Rs).unsqueeze(0), torch.mean(Rs).unsqueeze(0), torch.mean(Rs).unsqueeze(0)]).unsqueeze(0))
                elif self.agg == "min":
                    features.append(torch.cat([torch.min(Rs).unsqueeze(0), torch.min(Rs).unsqueeze(0), torch.min(Rs).unsqueeze(0)]).unsqueeze(0))
                elif self.agg == "max":
                    features.append(torch.cat([torch.max(Rs).unsqueeze(0), torch.max(Rs).unsqueeze(0), torch.max(Rs).unsqueeze(0)]).unsqueeze(0))
                elif self.agg == "all":
                    features.append(torch.cat([torch.min(Rs).unsqueeze(0), torch.mean(Rs).unsqueeze(0), torch.max(Rs).unsqueeze(0)]).unsqueeze(0))
            else:
                features.append(torch.FloatTensor([0.0, 0.0, 0.0]).unsqueeze(0)) # .cuda()
        features = torch.cat(features)
        logits = self.layer_final(features)
        histograms_out = [histogram.cpu().numpy() for histogram in histograms]
        return logits, histograms_out, images

    def save_imager_cache(self):
        for imager in self.imagers:
            imager.save_cache()

    def score(self, originals, generateds, **kwargs):
        with torch.no_grad():
            logits, histograms, images = self.forward(originals, generateds)
            probs = torch.nn.functional.softmax(logits, dim=-1)
            batch_scores = probs[:, 1].tolist()
        return {"scores": batch_scores} # , "histograms": histograms, "images": images


class SummaCZS:
    def __init__(self, model_name="mnli", granularity="paragraph", op1="max", op2="mean", use_ent=True, use_con=True, imager_load_cache=True, device="cuda", **kwargs):
        assert op2 in ["min", "mean", "max"], "Unrecognized `op2`"
        assert op1 in ["max", "mean", "min"], "Unrecognized `op1`"
        self.device = device
        self.imager = SummaCImager(model_name=model_name, granularity=granularity, device=self.device, **kwargs)
        if imager_load_cache:
            self.imager.load_cache()
        self.op2 = op2
        self.op1 = op1
        self.use_ent = use_ent
        self.use_con = use_con

    def save_imager_cache(self):
        self.imager.save_cache()

    def score_one(self, original, generated):
        image = self.imager.build_image(original, generated)
        score = self.image2score(image)
        return {"image": image, "score": score}

    def image2score(self, image):
        ent_scores = np.max(image[0], axis=0)
        co_scores = np.max(image[1], axis=0)
        if self.op1 == "mean":
            ent_scores = np.mean(image[0], axis=0)
            co_scores = np.mean(image[1], axis=0)
        elif self.op1 == "min":
            ent_scores = np.min(image[0], axis=0)
            co_scores = np.min(image[1], axis=0)

        if self.use_ent and self.use_con:
            scores = ent_scores - co_scores
        elif self.use_ent:
            scores = ent_scores
        elif self.use_con:
            scores = 1.0 - co_scores

        final_score = np.mean(scores)
        if self.op2 == "min":
            final_score = np.min(scores)
        elif self.op2 == "max":
            final_score = np.max(scores)
        return final_score

    def score(self, sources, generateds, batch_size=128, **kwargs):
        images = self.imager.build_images(sources, generateds, batch_size=batch_size)
        scores = [self.image2score(image) for image in images]
        return {"scores": scores, "images": images}


if __name__ == "__main__":
    model = SummaCZS(granularity="document", model_name="vitc", imager_load_cache=True, device="cpu") # Device can be `cpu` or `cuda` when GPU is available

    document = "Jeff joined Microsoft in 1992 to lead corporate developer evangelism for Windows NT."
    summary1 = "Jeff joined Microsoft in 1992."
    summary2 = "Jeff joined Microsoft."

    print(model.score([document, document], [summary1, summary2])["scores"])

    # document = """Jeff joined Microsoft in 1992 to lead corporate developer evangelism for Windows NT. He then served as a Group Program manager in Microsoft's Internet Business Unit. In 1998, he led the creation of SharePoint Portal Server, which became one of Microsoft’s fastest-growing businesses, exceeding $2 billion in revenues. Jeff next served as Corporate Vice President for Program Management across Office 365 Services and Servers, which is the foundation of Microsoft's enterprise cloud leadership. He then led Corporate Strategy supporting Satya Nadella and Amy Hood on Microsoft's mobile-first/cloud-first transformation and acquisitions. Prior to joining Microsoft, Jeff was vice president for software development for an investment firm in New York. He leads Office shared experiences and core applications, as well as OneDrive and SharePoint consumer and business services in Office 365. Jeff holds a Master of Business Administration degree from Harvard Business School and a Bachelor of Science degree in information systems and finance from New York University."""
    # summary = "Jeff joined Microsoft in 1992 to lead the company's corporate evangelism. He then served as a Group Manager in Microsoft's Internet Business Unit. In 1998, Jeff led Sharepoint Portal Server, which became the company's fastest-growing business, surpassing $3 million in revenue. Jeff next leads corporate strategy for SharePoint and Servers which is the basis of Microsoft's cloud-first strategy. He leads corporate strategy for Satya Nadella and Amy Hood on Microsoft's mobile-first."

    # scores = model.score([document], [summary])["images"][0][0].T
    # summary_sentences = model.imager.split_text(summary)

    # print(np.array2string(scores, precision=2))
    # for score_row, sentence in zip(scores, summary_sentences):
    #     print("-----------")
    #     print("[SummaC score: %.3f; supporting sentence: %d] %s " % (np.max(score_row), np.argmax(score_row)+1, sentence))



# borrow code from https://github.com/timoschick/self-debiasing

from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from transformers.file_utils import ModelOutput

import torch

@dataclass
class SampleEncoderDecoderOutput(ModelOutput):
    """
    Base class for outputs of encoder-decoder generation models using sampling. Hidden states and attention weights of
    the decoder (respectively the encoder) can be accessed via the encoder_attentions and the encoder_hidden_states
    attributes (respectively the decoder_attentions and the decoder_hidden_states attributes)


    Args:
        sequences (:obj:`torch.LongTensor` of shape :obj:`(batch_size*num_return_sequences, sequence_length)`):
            The generated sequences. The second dimension (sequence_length) is either equal to :obj:`max_length` or
            shorter if all batches finished early due to the :obj:`eos_token_id`.
        scores (:obj:`tuple(torch.FloatTensor)` `optional`, returned when ``output_scores=True`` is passed or when ``config.output_scores=True``):
            Processed prediction scores of the language modeling head (scores for each vocabulary token before SoftMax)
            at each generation step. :obj:`(max_length,)`-shaped tuple of :obj:`torch.FloatTensor` with each tensor of
            shape :obj:`(batch_size*num_return_sequences, config.vocab_size)`).
        encoder_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer of the decoder) of shape
            :obj:`(batch_size*num_return_sequences, num_heads, sequence_length, sequence_length)`.
        encoder_hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size*num_return_sequences, sequence_length, hidden_size)`.
        decoder_attentions (:obj:`tuple(tuple(torch.FloatTensor))`, `optional`, returned when ``output_attentions=True`` is passed or ``config.output_attentions=True``):
            Tuple (one element for each generated token) of tuples (one element for each layer of the decoder) of
            :obj:`torch.FloatTensor` of shape :obj:`(batch_size*num_return_sequences, num_heads, generated_length,
            sequence_length)`.
        decoder_hidden_states (:obj:`tuple(tuple(torch.FloatTensor))`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple (one element for each generated token) of tuples (one element for each layer of the decoder) of
            :obj:`torch.FloatTensor` of shape :obj:`(batch_size*num_return_sequences, generated_length, hidden_size)`.
    """

    sequences: torch.LongTensor = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    encoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    encoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    decoder_attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    decoder_hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None


@dataclass
class SampleDecoderOnlyOutput(ModelOutput):
    """
    Base class for outputs of decoder-only generation models using sampling.


    Args:
        sequences (:obj:`torch.LongTensor` of shape :obj:`(batch_size*num_return_sequences, sequence_length)`):
            The generated sequences. The second dimension (sequence_length) is either equal to :obj:`max_length` or
            shorter if all batches finished early due to the :obj:`eos_token_id`.
        scores (:obj:`tuple(torch.FloatTensor)` `optional`, returned when ``output_scores=True`` is passed or when ``config.output_scores=True``):
            Processed prediction scores of the language modeling head (scores for each vocabulary token before SoftMax)
            at each generation step. :obj:`(max_length,)`-shaped tuple of :obj:`torch.FloatTensor` with each tensor of
            shape :obj:`(batch_size*num_return_sequences, config.vocab_size)`).
        attentions (:obj:`tuple(tuple(torch.FloatTensor))`, `optional`, returned when ``output_attentions=True`` is passed or ``config.output_attentions=True``):
            Tuple (one element for each generated token) of tuples (one element for each layer of the decoder) of
            :obj:`torch.FloatTensor` of shape :obj:`(num_return_sequences*batch_size, num_heads, generated_length,
            sequence_length)`.
        hidden_states (:obj:`tuple(tuple(torch.FloatTensor))`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple (one element for each generated token) of tuples (one element for each layer of the decoder) of
            :obj:`torch.FloatTensor` of shape :obj:`(num_return_sequences*batch_size, generated_length, hidden_size)`.
    """

    sequences: torch.LongTensor = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None

SampleOutput = Union[SampleEncoderDecoderOutput, SampleDecoderOnlyOutput]


# Borrowed from https://github.com/maszhongming/UniEval/blob/main/metric/evaluator.py

from .scorer import UniEvaluator
from ..utils import retrieve_global_knowledge, get_instruction
from nltk import sent_tokenize

class FactEvaluator:
    def __init__(self, max_length=1024, device='cuda:0', cache_dir=None):
        """ Set up evaluator for factual consistency detection """
        self.scorer = UniEvaluator(model_name_or_path='MingZhong/unieval-fact',
                                   max_length=max_length,
                                   device=device, cache_dir=cache_dir)
        self.task = 'fact'
        self.dim = 'consistency'

    def evaluate(self, data, global_knowledge):
        """
            Get the factual consistency score (only 1 dimension for this task)

            print_result: whether to print the average factual score on the screen
        """
        n_data = len(data)

        # Calculate average sentence-level scores for facutal consistency
        input_list = []
        n_sents = [] # the number of sentences in the claim
        for sample in data:
            source = '\n'.join([get_instruction(sample),
                                sample['knowledge'],
                                retrieve_global_knowledge(sample['output'], global_knowledge)])
            system_outputs = sent_tokenize(sample['output'])
            n_sents.append(len(system_outputs))
            for system_output in system_outputs:
                input_list.append('question: Is this claim consistent with the document? </s> claim: ' + system_output + ' </s> document: ' + source)
        sent_score = self.scorer.score(input_list)

        # Get average score for each sample
        start_idx = 0
        scores = []
        for cur_n_sent in n_sents:
            scores.append(sum(sent_score[start_idx: start_idx + cur_n_sent]) / cur_n_sent)
            start_idx += cur_n_sent

        return scores
